/*
  CRF++ -- Yet Another CRF toolkit

  $Id: feature_index.cpp,v 1.15 2006/03/28 00:26:54 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#include <iostream>
#include <fstream>
#include <cstring>
#include <set>
#include "common.h"
#include "feature_index.h"

namespace CRFPP
{
  static inline char *read_ptr (char **ptr, size_t size)
  {
    char *r = *ptr;
    *ptr += size;
    return r;
  }

  template <class T> static inline void read_static (char **ptr, T& value)
  {
    char *r = read_ptr (ptr, sizeof (T));
    memcpy (&value, r, sizeof (T));
  }

  int DecoderFeatureIndex::getID (const char *key)
  {
    return da.exactMatchSearch<Darts::DoubleArray::result_type>(key);
  }

  int EncoderFeatureIndex::getID (const char *key)
  {
    std::map <std::string, std::pair<int,unsigned int> >::iterator it = dic.find (key);
    if (it == dic.end()) {
      dic.insert (std::make_pair<std::string, std::pair<int, unsigned int> >
		  (key, std::make_pair<int, unsigned int>(maxid, 1)));
      int n = maxid;
      maxid += (key[0] == 'U' ? y.size() : y.size() * y.size());
      return n;
    } else {
      it->second.second++;
      return it->second.first;
    }
    return -1;
  }

  bool EncoderFeatureIndex::open (const char *filename1,
				  const char *filename2)
  {
    return openTemplate(filename1) && openTagSet(filename2);
  }

  bool EncoderFeatureIndex::openTemplate (const char *filename)
  {
    std::ifstream ifs (filename);
    if (!ifs)
      WHAT_FALSE("open failed: "  << filename) ;

    std::string line;
    while (std::getline (ifs, line)) {
      if (! line[0] || line[0] == '#') continue;
      if (line[0] == 'U') {
	unigram_templs.push_back(this->strdup(line.c_str()));
      } else if (line[0] == 'B') {
	bigram_templs.push_back(this->strdup (line.c_str()));
      } else {
	WHAT_FALSE("unknown type: " << line << " " << filename);
      }
    }

    return true;
  }

  bool EncoderFeatureIndex::openTagSet (const char *file)
  {
    std::ifstream ifs (file);
    if (! ifs)	WHAT_FALSE("open failed: " << file) ;

    char  line[8192];
    char* column[1024];
    size_t max_size = 0;
    std::set<std::string> candset;

    while (ifs.getline (line, sizeof(line))) {
      if (line[0] == '\0' || line[0] == ' ' || line[0] == '\t') continue;
      size_t _size = tokenize (line, "\t ", column, 1024);
      if (max_size == 0) max_size = _size;
      else if (max_size != _size)
	WHAT_FALSE("inconsistent column size: " << max_size << " " << _size << " " << file);

      xsize_ = _size - 1;
      candset.insert(column[max_size-1]);
    }

    y.clear ();
    for (std::set<std::string>::iterator it = candset.begin(); it != candset.end(); ++it)
      y.push_back (this->strdup (it->c_str()));

    ifs.close();

    return true;
  }

  bool DecoderFeatureIndex::open(const char *filename1,
				 const char *filename2)
  {
    if (! mmap.open (filename1)) WHAT_FALSE(mmap.what ()) ;

    char *ptr = mmap.begin ();
    unsigned int version_ = 0;

    read_static<unsigned int>(&ptr, version_);

    if (version_ / 100 != version / 100)
      WHAT_FALSE("model version is different: " << version_ 
		 << " vs " << version << " : " << filename1);

    int type = 0;
    read_static<int>(&ptr, type);
    double cost_factor = 0.0;
    read_static<double>(&ptr, cost_factor);

    read_static<unsigned int>(&ptr, maxid);
    read_static<unsigned int>(&ptr, xsize_);

    unsigned int dsize = 0;
    read_static<unsigned int>(&ptr, dsize);

    unsigned int y_str_size;
    read_static<unsigned int>(&ptr, y_str_size);
    char *y_str = read_ptr(&ptr, y_str_size);
    size_t pos = 0;
    while (pos < y_str_size) {
      y.push_back (y_str + pos);
      while (y_str[pos++] != '\0') {};
    }

    unsigned int tmpl_str_size;
    read_static<unsigned int>(&ptr, tmpl_str_size);
    char *tmpl_str = read_ptr(&ptr, tmpl_str_size);
    pos = 0;
    while (pos < tmpl_str_size) {
      char *v = tmpl_str + pos;
      if (v[0] == '\0') {
	++pos;
      } else if (v[0] == 'U') {
	unigram_templs.push_back (v);
      } else if (v[0] == 'B') {
	bigram_templs.push_back (v);
      } else {
	WHAT("unknown type: " << v);
      }
      while (tmpl_str[pos++] != '\0') {};
    }

    da.set_array (ptr);
    ptr += dsize;

    alpha_float = (float *)ptr;
    ptr += sizeof (float) * maxid;

    if (ptr != mmap.end())
      WHAT_FALSE("model file is broken: " << filename1);

    return true;
  }

  void EncoderFeatureIndex::shrink (size_t freq)
  {
    if (freq <= 1) return;

    std::map<int,int> old2new;
    int new_maxid = 0;

    for (std::map<std::string, std::pair<int, unsigned int> >::iterator
	   it = dic.begin(); it != dic.end();) {
      const std::string &key = it->first;

      if (it->second.second >= freq) {
	old2new.insert(std::make_pair<int,int>(it->second.first, new_maxid));
	it->second.first = new_maxid;
	new_maxid += (key[0] == 'U' ? y.size() : y.size() * y.size());
	++it;
      } else {
	dic.erase (it++);
      }
    }

    feature_cache.shrink (old2new);

    maxid = new_maxid;

    return;
  }

  void DecoderFeatureIndex::clear ()
  {
    char_freelist.free();
    feature_cache.clear();
    for (size_t i = 0; i < thread_num; ++i) {
      node_freelist[i].free();
      path_freelist[i].free();
    }
  }

  void EncoderFeatureIndex::clear () {}

  bool EncoderFeatureIndex::convert(const char *filename1,
				    const char *filename2)
  {
    std::ifstream ifs (filename1);

    y.clear();
    dic.clear();
    unigram_templs.clear();
    bigram_templs.clear();
    xsize_ = 0;
    maxid = 0;

    if (!ifs) 
      WHAT_FALSE("open failed: " << filename1);

    char line[8192];
    char *column[8];

    // read header
    while (true) {
      if (! ifs.getline(line, sizeof(line)))
	WHAT_FALSE("format error: " << filename1);

      if (std::strlen(line) == 0) break;

      size_t _size = tokenize(line, "\t ", column, 2);

      if (_size != 2) 
	WHAT_FALSE("format error: " << filename1);

      if (std::strcmp(column[0], "xsize:") == 0) 
	xsize_ = std::atoi(column[1]);

      if (std::strcmp(column[0], "maxid:") == 0) 
	maxid = std::atoi(column[1]);
    }

    if (maxid == 0) 
      WHAT_FALSE("maxid is not defined: " << filename1);

    if (xsize_ == 0) 
      WHAT_FALSE("xsize is not defined: " << filename1);

    while (true) {
      if (! ifs.getline(line, sizeof(line)))
	WHAT_FALSE("format error: " << filename1);
      if (std::strlen(line) == 0) break;
      y.push_back(this->strdup(line));
    }

    while (true) {
      if (! ifs.getline(line, sizeof(line)))
	WHAT_FALSE("format error: " << filename1);
      if (std::strlen(line) == 0) break;
      if (line[0] == 'U') {
	unigram_templs.push_back(this->strdup(line));
      } else if (line[0] == 'B') {
	bigram_templs.push_back(this->strdup (line));
      } else {
	WHAT_FALSE("unknown type: " << line << " " << filename1);
      }
    }

    while (true) {
      if (! ifs.getline(line, sizeof(line)))
	WHAT_FALSE("format error: " << filename1);
      if (std::strlen(line) == 0) break;

      size_t _size = tokenize(line, "\t ", column, 2);
      if (_size != 2) 
	WHAT_FALSE("format error: " << filename1);

      dic.insert (std::make_pair<std::string, std::pair<int, unsigned int> >
		  (column[1], std::make_pair<int, unsigned int>(std::atoi(column[0]), 1)));
    }

    std::vector<double> alpha_;
    while (ifs.getline(line, sizeof(line))) 
      alpha_.push_back(std::atof(line));

    alpha = &alpha_[0];

    if (alpha_.size() != maxid) 
      WHAT_FALSE("maxid != alpha.size(): " << filename1);

    return save(filename2, false);
  }


  bool EncoderFeatureIndex::save (const char *filename, bool textmodelfile)
  {
    std::vector <char *> key;
    std::vector <int>	 val;

    std::string y_str;
    for (size_t i = 0; i < y.size(); ++i) {
      y_str += std::string (y[i]);
      y_str += '\0';
    }

    std::string templ_str;
    for (size_t i = 0; i < unigram_templs.size(); ++i) {
      templ_str += std::string (unigram_templs[i]);
      templ_str += '\0';
    }

    for (size_t i = 0; i < bigram_templs.size(); ++i) {
      templ_str += std::string (bigram_templs[i]);
      templ_str += '\0';
    }

    while ((y_str.size() + templ_str.size()) % 4 != 0) templ_str += '\0';

    for (std::map<std::string, std::pair<int, unsigned int> >::iterator it = dic.begin();
	 it != dic.end(); ++it) {
      key.push_back ((char *)(it->first.c_str()));
      val.push_back (it->second.first);
    }

    Darts::DoubleArray da;

    if (da.build (key.size(), &key[0], 0, &val[0]) < 0)
      WHAT_FALSE("cannot build double-array");

    std::ofstream bofs;
    bofs.open(filename, OUTPUT_MODE);

    if (! bofs) 
      WHAT_FALSE("open failed: " << filename);

    unsigned int version_ = version;
    bofs.write ((char *)(&version_),sizeof (unsigned int));

    int type = 0;
    bofs.write ((char *)(&type),sizeof (int));

    double cost_factor = 1.0;
    bofs.write ((char *)(&cost_factor),sizeof (double));

    bofs.write ((char *)(&maxid),  sizeof (unsigned int));
    bofs.write ((char *)(&xsize_), sizeof (unsigned int));

    unsigned int dsize = da.unit_size () * da.size();
    bofs.write ((char *)(&dsize), sizeof (unsigned int));

    unsigned int size = y_str.size();
    bofs.write ((char *)(&size),  sizeof (unsigned int));
    bofs.write ((char *)y_str.data(),	 y_str.size());

    size = templ_str.size();
    bofs.write ((char *)(&size),  sizeof (unsigned int));
    bofs.write ((char *)templ_str.data(), templ_str.size());

    bofs.write ((char *)da.array(), dsize);

    for (size_t i  = 0; i < maxid; ++i) {
      float alpha_ = (float)alpha[i];
      bofs.write ((char *)(&alpha_), sizeof (float));
    }

    bofs.close ();

    if (textmodelfile) {

      std::string filename2 = filename;
      filename2 += ".txt";

      std::ofstream tofs (filename2.c_str());

      if (! tofs) 
        WHAT_FALSE("open failed: " << filename2);

      // header
      tofs << "version: "     << version_ << std::endl;
      tofs << "cost-factor: " << cost_factor << std::endl;
      tofs << "maxid: "       << maxid << std::endl;
      tofs << "xsize: "       << xsize_ << std::endl;

      tofs << std::endl;

      // y
      for (size_t i = 0; i < y.size(); ++i) 
        tofs << y[i] << std::endl;

      tofs << std::endl;

      // template
      for (size_t i = 0; i < unigram_templs.size(); ++i) 
        tofs << unigram_templs[i] << std::endl;
    
      for (size_t i = 0; i < bigram_templs.size(); ++i) 
        tofs << bigram_templs[i] << std::endl;

      tofs << std::endl;

      // dic
      for (std::map<std::string, std::pair<int, unsigned int> >::iterator it = dic.begin();
           it != dic.end(); ++it) {
        tofs << it->second.first << " " << it->first << std::endl;
      }

      tofs << std::endl;

      tofs.setf(std::ios::fixed, std::ios::floatfield);
      tofs.precision(16);

      for (size_t i  = 0; i < maxid; ++i) 
        tofs << alpha[i] << std::endl;
    }

    return true;
  }

  char *FeatureIndex::strdup (const char *p)
  {
    size_t len = std::strlen (p);
    char *q = char_freelist.alloc (len+1);
    std::strcpy (q, p);
    return q;
  }

  void FeatureIndex::calcCost (Node *n)
  {
    n->cost = 0.0;

#define ADD_COST(T, A) \
      { T c = 0; \
	for (int *f = n->fvector; *f != -1; ++f) c += (A)[*f + n->y]; \
	n->cost = (T)c; } 

    if (alpha_float) ADD_COST(float,  alpha_float)
		       else		ADD_COST(double, alpha);
#undef ADD_COST
  }

  void FeatureIndex::calcCost (Path *p)
  {
    p->cost = 0.0;

#define ADD_COST(T, A) \
      { T c = 0.0; \
	for (int *f = p->fvector; *f != -1; ++f) \
	  c += (A)[*f + p->lnode->y * y.size() + p->rnode->y]; \
	p->cost = (T)c; }

    if (alpha_float) ADD_COST(float,  alpha_float)
		       else		ADD_COST(double, alpha);
  }
#undef ADD_COST
}
